var Connection, Document, MicroEvent, connections, getConnection, io, open, types;
var __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };
if (typeof WEB !== "undefined" && WEB !== null) {
  types || (types = exports.types);
  if (!window['io']) {
    throw new Error('Must load socket.io before this library');
  }
  io = window['io'];
} else {
  types = require('../types');
  io = require('../../thirdparty/Socket.io-node-client').io;
  MicroEvent = require('./microevent');
}
Document = function(connection, name, version, type, snapshot) {
  var inflightCallbacks, inflightOp, k, listeners, pendingCallbacks, pendingOp, serverOps, tryFlushPendingOp, v, _ref;
  this.name = name;
  this.version = version;
  this.type = type;
  if (this.type.compose == null) {
    throw new Error('Handling types without compose() defined is not currently implemented');
  }
  this.setSnapshot = function(s) {
    return this['snapshot'] = this.snapshot = s;
  };
  this.setSnapshot(snapshot);
  inflightOp = null;
  inflightCallbacks = [];
  pendingOp = null;
  pendingCallbacks = [];
  serverOps = {};
  listeners = [];
  tryFlushPendingOp = __bind(function() {
    if (inflightOp === null && pendingOp !== null) {
      inflightOp = pendingOp;
      inflightCallbacks = pendingCallbacks;
      pendingOp = null;
      pendingCallbacks = [];
      return connection.send({
        'doc': this.name,
        'op': inflightOp,
        'v': this.version
      }, __bind(function(response) {
        var callback, _i, _j, _len, _len2;
        if (response['v'] === null) {
          for (_i = 0, _len = inflightCallbacks.length; _i < _len; _i++) {
            callback = inflightCallbacks[_i];
            callback(null);
          }
          inflightOp = null;
          throw new Error(response['error']);
        }
        if (response['v'] !== this.version) {
          throw new Error('Invalid version from server');
        }
        serverOps[this.version] = inflightOp;
        this.version++;
        for (_j = 0, _len2 = inflightCallbacks.length; _j < _len2; _j++) {
          callback = inflightCallbacks[_j];
          callback(inflightOp, null);
        }
        inflightOp = null;
        return tryFlushPendingOp();
      }, this));
    }
  }, this);
  this._onOpReceived = function(msg) {
    var docOp, oldSnapshot, op, xf, _ref, _ref2;
    if (msg['v'] < this.version) {
      return;
    }
    if (msg['doc'] !== this.name) {
      throw new Error("Expected docName " + this.name + " but got " + msg['doc']);
    }
    if (msg['v'] !== this.version) {
      throw new Error("Expected version " + this.version + " but got " + msg['v']);
    }
    op = msg['op'];
    serverOps[this.version] = op;
    xf = this.type.transformX || __bind(function(client, server) {
      var client_, server_;
      client_ = this.type.transform(client, server, 'left');
      server_ = this.type.transform(server, client, 'right');
      return [client_, server_];
    }, this);
    docOp = op;
    if (inflightOp !== null) {
      _ref = xf(inflightOp, docOp), inflightOp = _ref[0], docOp = _ref[1];
    }
    if (pendingOp !== null) {
      _ref2 = xf(pendingOp, docOp), pendingOp = _ref2[0], docOp = _ref2[1];
    }
    oldSnapshot = this.snapshot;
    this.setSnapshot(this.type.apply(oldSnapshot, docOp));
    this.version++;
    this.emit('remoteop', docOp, oldSnapshot);
    return this.emit('change', docOp, oldSnapshot);
  };
  this['submitOp'] = this.submitOp = function(op, v, callback) {
    var realOp, _ref;
    if (v == null) {
      v = this.version;
    }
    if (typeof v === 'function') {
      callback = v;
      v = this.version;
    }
    if (((_ref = this.type) != null ? _ref.normalize : void 0) != null) {
      op = this.type.normalize(op);
    }
    while (v < this.version) {
      realOp = serverOps[v];
      if (!realOp) {
        throw new Error('Op version too old');
      }
      op = this.type.transform(op, realOp, 'left');
      v++;
    }
    this.setSnapshot(this.type.apply(this.snapshot, op));
    if (pendingOp !== null) {
      pendingOp = this.type.compose(pendingOp, op);
    } else {
      pendingOp = op;
    }
    if (callback) {
      pendingCallbacks.push(callback);
    }
    this.emit('change', op);
    return setTimeout(tryFlushPendingOp, 0);
  };
  this['close'] = this.close = function(callback) {
    return connection.send({
      'doc': this.name,
      open: false
    }, __bind(function() {
      if (callback) {
        callback();
      }
      this.emit('closed');
    }, this));
  };
  if (this.type.api) {
    _ref = this.type.api;
    for (k in _ref) {
      v = _ref[k];
      this[k] = v;
    }
    this._register();
  } else {
    this.provides = this['provides'] = {};
  }
  return this;
};
MicroEvent.mixin(Document);
Connection = (function() {
  function Connection(host, port, basePath) {
    this.onMessage = __bind(this.onMessage, this);
    this.connected = __bind(this.connected, this);
    this.disconnected = __bind(this.disconnected, this);
    var resource;
    resource = basePath ? path + '/socket.io' : 'socket.io';
    this.socket = new io['Socket'](host, {
      port: port,
      resource: resource
    });
    this.socket['on']('connect', this.connected);
    this.socket['on']('disconnect', this.disconnected);
    this.socket['on']('message', this.onMessage);
    this.socket['connect']();
    this.docs = {};
    this.numDocs = 0;
  }
  Connection.prototype.disconnected = function() {
    return this.emit('disconnect');
  };
  Connection.prototype.connected = function() {
    return this.emit('connect');
  };
  Connection.prototype.send = function(msg, callback) {
    var docName, register;
    docName = msg['doc'];
    if (docName === this.lastSentDoc) {
      delete msg['doc'];
    } else {
      this.lastSentDoc = docName;
    }
    this.socket['send'](msg);
    if (callback) {
      register = __bind(function(type) {
        var cb;
        cb = __bind(function(response) {
          if (response['doc'] === docName) {
            this.removeListener(type, cb);
            return callback(response);
          }
        }, this);
        return this.on(type, cb);
      }, this);
      return register((msg['open'] === true ? 'open' : msg['open'] === false ? 'close' : msg['create'] ? 'create' : msg['snapshot'] === null ? 'snapshot' : msg['op'] ? 'op response' : void 0));
    }
  };
  Connection.prototype.onMessage = function(msg) {
    var doc, docName, type;
    docName = msg['doc'];
    if (docName !== void 0) {
      this.lastReceivedDoc = docName;
    } else {
      msg['doc'] = docName = this.lastReceivedDoc;
    }
    this.emit('message', msg);
    type = msg['open'] === true || (msg['open'] === false && msg['error']) ? 'open' : msg['open'] === false ? 'close' : msg['snapshot'] !== void 0 ? 'snapshot' : msg['create'] ? 'create' : msg['op'] ? 'op' : msg['v'] !== void 0 ? 'op response' : void 0;
    this.emit(type, msg);
    if (type === 'op') {
      doc = this.docs[docName];
      if (doc) {
        return doc._onOpReceived(msg);
      }
    }
  };
  Connection.prototype.makeDoc = function(params) {
    var doc, name, type;
    name = params['doc'];
    if (this.docs[name]) {
      throw new Error("Document " + name + " already followed");
    }
    type = params['type'];
    if (typeof type === 'string') {
      type = types[type];
    }
    doc = new Document(this, name, params['v'], type, params['snapshot']);
    doc['created'] = !!params['create'];
    this.docs[name] = doc;
    this.numDocs++;
    doc.on('closed', __bind(function() {
      delete this.docs[name];
      return this.numDocs--;
    }, this));
    return doc;
  };
  Connection.prototype['openExisting'] = function(docName, callback) {
    if (this.docs[docName] != null) {
      return this.docs[docName];
    }
    return this.send({
      'doc': docName,
      'open': true,
      'snapshot': null
    }, __bind(function(response) {
      if (response.error) {
        return callback(null, new Error(response.error));
      } else {
        return callback(this.makeDoc(response));
      }
    }, this));
  };
  Connection.prototype['open'] = function(docName, type, callback) {
    var doc;
    if (typeof type === 'function') {
      callback = type;
      type = 'text';
    }
    callback || (callback = function() {});
    if (typeof type === 'string') {
      type = types[type];
    }
    if ((docName != null) && (this.docs[docName] != null)) {
      doc = this.docs[docName];
      if (doc.type === type) {
        callback(doc);
      } else {
        callback(doc, 'Type mismatch');
      }
      return;
    }
    return this.send({
      'doc': docName,
      'open': true,
      'create': true,
      'snapshot': null,
      'type': type.name
    }, __bind(function(response) {
      if (response.error) {
        return callback(null, response.error);
      } else {
        if (response['snapshot'] === void 0) {
          response['snapshot'] = type.create();
        }
        response['type'] = type;
        return callback(this.makeDoc(response));
      }
    }, this));
  };
  Connection.prototype['create'] = function(type, callback) {
    return open(null, type, callback);
  };
  Connection.prototype['disconnect'] = function() {
    if (this.stream != null) {
      this.emit('disconnected');
      this.stream.disconnect();
      return this.stream = null;
    }
  };
  return Connection;
})();
MicroEvent.mixin(Connection);
connections = {};
getConnection = function(host, port, basePath) {
  var address, c;
  if (typeof WEB !== "undefined" && WEB !== null) {
    if (host == null) {
      host = window.location.hostname;
    }
    if (port == null) {
      port = window.location.port;
    }
  }
  address = host;
  if (port != null) {
    address += ":" + port;
  }
  if (!connections[address]) {
    c = new Connection(host, port, basePath);
    c.on('disconnected', function() {
      return delete connections[address];
    });
    connections[address] = c;
  }
  return connections[address];
};
open = function(docName, type, options, callback) {
  var c;
  if (typeof options === 'function') {
    callback = options;
    options = null;
  }
  if (options == null) {
    options = {};
  }
  c = getConnection(options.host, options.port, options.basePath);
  return c.open(docName, type, function(doc, error) {
    if (doc === null) {
      return callback(null, error);
    } else {
      doc.on('closed', function() {
        return setTimeout(function() {
          if (c.numDocs === 0) {
            return c.disconnect();
          }
        }, 0);
      });
      return callback(doc);
    }
  });
};
if (typeof WEB !== "undefined" && WEB !== null) {
  exports['Connection'] = Connection;
  exports['Document'] = Document;
  exports['open'] = open;
  window['sharejs'] = exports;
} else {
  exports.Connection = Connection;
  exports.open = open;
}