var Events, Model, db, i, p, queue, types;
var __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };
p = function() {};
i = function() {};
queue = require('./syncqueue');
types = require('../types');
db = require('./db');
Events = require('./events');
module.exports = Model = function(db, options) {
  var auth, events, getSnapshot, queues;
  if (!(this instanceof Model)) {
    return new Model(db);
  }
  if (options == null) {
    options = {};
  }
  this.getOps = function(docName, start, end, callback) {
    return db.getVersion(docName, function(v) {
      if (v === null) {
        return callback(null, 'Document does not exist');
      } else {
        return db.getOps(docName, start, end, callback);
      }
    });
  };
  this.getSnapshot = getSnapshot = function(docName, callback) {
    return db.getSnapshot(docName, function(data, error) {
      p("getSnapshot " + (i(data)));
      if (data != null) {
        if (data != null ? data.type : void 0) {
          data.type = types[data.type];
        }
        return callback(data);
      } else {
        return callback(data, error);
      }
    });
  };
  this.getVersion = db.getVersion;
  this.create = function(docName, type, meta, callback) {
    var newDocData;
    if (typeof type === 'string') {
      type = types[type];
    }
    if (typeof meta === 'function') {
      callback = meta;
      meta = {};
    }
    meta || (meta = {});
    newDocData = {
      snapshot: type.create(),
      type: type.name,
      meta: meta || {},
      v: 0
    };
    p("db.create " + docName + ", " + (i(newDocData)));
    return db.create(docName, newDocData, callback);
  };
  queues = {};
  this.applyOp = function(docName, opData, callback) {
    p("applyOp " + docName + " op " + (i(opData)));
    queues[docName] || (queues[docName] = queue(function(opData, callback) {
      p("applyOpInternal v" + opData.v + " " + (i(opData.op)) + " to " + docName + ".");
      return getSnapshot(docName, function(docData) {
        var meta, op, opVersion, snapshot, submit, type, version;
        if (!docData) {
          callback(null, 'Document does not exist');
          return;
        }
        opVersion = opData.v;
        op = opData.op;
        meta = opData.meta || {};
        meta.ts = Date.now();
        version = docData.v, snapshot = docData.snapshot, type = docData.type;
        p("applyOp hasdata v" + opVersion + " " + (i(op)) + " to " + docName + ".");
        submit = function() {
          var newDocData, newOpData;
          try {
            snapshot = docData.type.apply(docData.snapshot, op);
          } catch (error) {
            callback(null, error.message);
            return;
          }
          newOpData = {
            op: op,
            v: opVersion,
            meta: meta
          };
          newDocData = {
            snapshot: snapshot,
            type: type.name,
            v: opVersion + 1,
            meta: docData.meta
          };
          p("submit " + (i(newOpData)));
          return db.append(docName, newOpData, newDocData, function() {
            p("appended v" + opVersion + " to " + docName + ". Calling callback...");
            events.onApplyOp(docName, newOpData);
            return callback(opVersion, void 0);
          });
        };
        if (opVersion > version) {
          callback(null, 'Op at future version');
          return;
        }
        if (opVersion < version) {
          return db.getOps(docName, opVersion, version, function(ops) {
            var realOp, _i, _len;
            try {
              for (_i = 0, _len = ops.length; _i < _len; _i++) {
                realOp = ops[_i];
                p("XFORM Doc " + docName + " op " + (i(op)) + " by " + (i(realOp.op)));
                op = docData.type.transform(op, realOp.op, 'left');
                opVersion++;
                p("-> " + (i(op)));
              }
            } catch (error) {
              callback(null, error.message);
              return;
            }
            return submit();
          });
        } else {
          return submit();
        }
      });
    }));
    return queues[docName](opData, callback);
  };
  this["delete"] = function(docName, callback) {
    events.removeAllListeners(docName);
    return db["delete"](docName, callback);
  };
  events = new Events(this);
  this.listen = events.listen;
  this.removeListener = events.removeListener;
  this.listenFromVersion = events.listenFromVersion;
  this.randomDocName = function(length) {
    var chars, x;
    if (length == null) {
      length = 10;
    }
    chars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-=";
    return ((function() {
      var _results;
      _results = [];
      for (x = 0; 0 <= length ? x < length : x > length; 0 <= length ? x++ : x--) {
        _results.push(chars[Math.floor(Math.random() * chars.length)]);
      }
      return _results;
    })()).join('');
  };
  options.auth || (options.auth = {});
  auth = options.auth || {};
  auth.auth || (auth.auth = function() {});
  auth.canCreate || (auth.canCreate = function(client, docName, type, meta, result) {
    return result.accept();
  });
  auth.canRead || (auth.canRead = function(client, docName, result) {
    return result.accept();
  });
  auth.canSubmitOp || (auth.canSubmitOp = function(client, docName, opData, result) {
    return result.accept();
  });
  auth.canDelete || (auth.canDelete = function(client, docName, result) {
    return result.reject();
  });
  this.clientGetOps = function(client, docName, start, end, callback) {
    return auth.canRead(client, docName, {
      accept: __bind(function() {
        return this.getOps(docName, start, end, callback);
      }, this),
      reject: function() {
        return callback(null, 'Forbidden');
      }
    });
  };
  this.clientGetSnapshot = function(client, docName, callback) {
    return auth.canRead(client, docName, {
      accept: __bind(function() {
        return this.getSnapshot(docName, callback);
      }, this),
      reject: function() {
        return callback(void 0, 'Forbidden');
      }
    });
  };
  this.clientCreate = function(client, docName, type, meta, callback) {
    if (typeof type === 'string') {
      type = types[type];
    }
    return auth.canCreate(client, docName, type, meta, {
      accept: __bind(function() {
        return this.create(docName, type, meta, callback);
      }, this),
      reject: function() {
        return callback(false, 'Forbidden');
      }
    });
  };
  this.clientSubmitOp = function(client, docName, opData, callback) {
    return auth.canSubmitOp(client, docName, opData, {
      accept: __bind(function() {
        return this.applyOp(docName, opData, callback);
      }, this),
      reject: function() {
        return callback(null, 'Forbidden');
      }
    });
  };
  this.clientDelete = function(client, docName, callback) {
    return auth.canDelete(client, docName, {
      accept: __bind(function() {
        return this["delete"](docName, callback);
      }, this),
      reject: function() {
        return callback(false, 'Forbidden');
      }
    });
  };
  this.clientListen = function(client, docName, listener, callback) {
    return auth.canRead(client, docName, {
      accept: __bind(function() {
        return this.listen(docName, listener, callback);
      }, this),
      reject: function() {
        return callback(null, 'Forbidden');
      }
    });
  };
  this.clientListenFromVersion = function(client, docName, version, listener, callback) {
    return auth.canRead(client, docName, {
      accept: __bind(function() {
        return this.listenFromVersion(docName, version, listener, callback);
      }, this),
      reject: function() {
        return callback(null, 'Forbidden');
      }
    });
  };
  return this;
};