var connect, expectJSONObject, http, pump, router, send200, send400, send404, sendJSON, sys, url, util;
http = require('http');
sys = require('sys');
util = require('util');
url = require('url');
connect = require('connect');
send404 = function(res, message) {
  if (message == null) {
    message = '404: Your document could not be found.\n';
  }
  res.writeHead(404, {
    'Content-Type': 'text/plain'
  });
  return res.end(message);
};
send400 = function(res, message) {
  res.writeHead(400, {
    'Content-Type': 'text/plain'
  });
  return res.end(message);
};
send200 = function(res, message) {
  if (message == null) {
    message = 'OK';
  }
  res.writeHead(200, {
    'Content-Type': 'text/plain'
  });
  return res.end(message);
};
sendJSON = function(res, obj) {
  res.writeHead(200, {
    'Content-Type': 'application/json'
  });
  return res.end(JSON.stringify(obj) + '\n');
};
expectJSONObject = function(req, res, callback) {
  return pump(req, function(data) {
    var obj;
    try {
      obj = JSON.parse(data);
      return callback(obj);
    } catch (error) {
      send400(res, 'Supplied JSON invalid');
    }
  });
};
pump = function(req, callback) {
  var data;
  data = '';
  req.on('data', function(chunk) {
    return data += chunk;
  });
  return req.on('end', function() {
    return callback(data);
  });
};
router = function(app, model, options) {
  app.get('/doc/:name', function(req, res) {
    return model.getSnapshot(req.params.name, function(doc) {
      if (doc) {
        res.setHeader('X-OT-Type', doc.type.name);
        res.setHeader('X-OT-Version', doc.v);
        if (typeof doc.snapshot === 'string') {
          return send200(res, doc.snapshot);
        } else {
          return sendJSON(res, doc.snapshot);
        }
      } else {
        return send404(res);
      }
    });
  });
  app.put('/doc/:name', function(req, res) {
    return expectJSONObject(req, res, function(obj) {
      var meta, type;
      type = obj != null ? obj.type : void 0;
      meta = obj != null ? obj.meta : void 0;
      if (!(typeof type === 'string' && (meta === void 0 || typeof meta === 'object'))) {
        return send400(res, 'Type invalid');
      } else {
        return model.create(req.params.name, type, meta, function(result, error) {
          if (result) {
            return send200(res);
          } else {
            return send400(res, error);
          }
        });
      }
    });
  });
  app.post('/doc/:name', function(req, res) {
    var query, version;
    query = url.parse(req.url, true).query;
    version = (query != null ? query.v : void 0) != null ? parseInt(query != null ? query.v : void 0) : parseInt(req.headers['x-ot-version']);
    if (!((version != null) && version >= 0)) {
      return send400(res, 'Version required - attach query parameter ?v=X on your URL or set the X-OT-Version header');
    } else {
      return expectJSONObject(req, res, function(obj) {
        var opData;
        opData = {
          v: version,
          op: obj,
          meta: {
            source: req.socket.remoteAddress
          }
        };
        return model.applyOp(req.params.name, opData, function(newVersion, error) {
          if (error != null) {
            return send400(res, error.stack);
          } else {
            return sendJSON(res, {
              v: newVersion
            });
          }
        });
      });
    }
  });
  if (options["delete"]) {
    return app["delete"]('/doc/:name', function(req, res) {
      return model["delete"](req.params.name, function(result) {
        if (result) {
          return send200(res);
        } else {
          return send404(res, "The document to be deleted doesn't exist");
        }
      });
    });
  }
};
module.exports = function(model, options) {
  var _ref;
  if (options == null) {
    options = {};
  }
  if ((_ref = options["delete"]) == null) {
    options["delete"] = false;
  }
  return connect.router(function(app) {
    return router(app, model, options);
  });
};